import os
import socket
import threading
import unittest
from sys import platform

import varlink

service = varlink.Service(
    vendor="Varlink",
    product="Varlink Examples",
    version="1",
    url="http://varlink.org",
    interface_dir=os.path.dirname(__file__),
)


class ServiceRequestHandler(varlink.RequestHandler):
    service = service


class TestService(unittest.TestCase):
    def do_run(self, address):
        server = varlink.ThreadingServer(address, ServiceRequestHandler)
        server_thread = threading.Thread(target=server.serve_forever)
        server_thread.daemon = True
        server_thread.start()

        try:
            with varlink.Client(address) as client, client.open("org.varlink.service") as _connection:
                info = _connection.GetInfo()

                self.assertEqual(len(info["interfaces"]), 1)
                self.assertEqual(info["interfaces"][0], "org.varlink.service")
                self.assertEqual(info, service.GetInfo())

                desc = _connection.GetInterfaceDescription(info["interfaces"][0])
                self.assertEqual(desc, service.GetInterfaceDescription("org.varlink.service"))

                _connection.close()

        finally:
            server.shutdown()
            server.server_close()

    def test_tcp(self) -> None:
        self.do_run("tcp:127.0.0.1:23450")

    def test_anon_unix(self) -> None:
        if platform.startswith("linux"):
            self.do_run(f"unix:@org.varlink.service_anon_test{os.getpid()}{threading.current_thread().name}")

    def test_unix(self) -> None:
        if hasattr(socket, "AF_UNIX"):
            self.do_run(f"unix:org.varlink.service_anon_test_{os.getpid()}{threading.current_thread().name}")

    def test_wrong_url(self) -> None:
        self.assertRaises(
            ConnectionError, self.do_run, f"uenix:org.varlink.service_wrong_url_test_{os.getpid()}"
        )

    def test_reuse_open(self) -> None:
        address = "tcp:127.0.0.1:23450"
        server = varlink.ThreadingServer(address, ServiceRequestHandler)
        server_thread = threading.Thread(target=server.serve_forever)
        server_thread.daemon = True
        server_thread.start()

        try:
            with varlink.Client(address) as client:
                connection = client.open_connection()
                re_use = client.open("org.varlink.service", False, connection)

                info = re_use.GetInfo()
                self.assertEqual(len(info["interfaces"]), 1)
                self.assertEqual(info["interfaces"][0], "org.varlink.service")
                self.assertEqual(info, service.GetInfo())
                connection.close()
        finally:
            server.shutdown()
            server.server_close()



import json
from types import SimpleNamespace


class VarlinkEncoder(json.JSONEncoder):
    """The Encoder used to encode JSON"""

    def default(self, o):
        if isinstance(o, set):
            return dict.fromkeys(o, {})
        if isinstance(o, SimpleNamespace):
            return o.__dict__
        if isinstance(o, VarlinkError):
            return o.as_dict()
        return json.JSONEncoder.default(self, o)


class VarlinkError(Exception):
    """The base class for varlink error exceptions"""

    @classmethod
    def new(cls, message, namespaced=False):
        if message["error"] == "org.varlink.service.InterfaceNotFound":
            return InterfaceNotFound.new(message, namespaced)

        elif message["error"] == "org.varlink.service.InvalidParameter":
            return InvalidParameter.new(message, namespaced)

        elif message["error"] == "org.varlink.service.MethodNotImplemented":
            return MethodNotImplemented.new(message, namespaced)

        elif message["error"] == "org.varlink.service.MethodNotImplemented":
            return MethodNotImplemented.new(message, namespaced)

        elif message["error"] == "org.varlink.service.MethodNotFound":
            return MethodNotFound.new(message, namespaced)

        else:
            return cls(message, namespaced)

    def __init__(self, message, namespaced=False):
        if not namespaced and not isinstance(message, dict):
            raise TypeError
        # normalize to dictionary
        Exception.__init__(self, json.loads(json.dumps(message, cls=VarlinkEncoder)))

    def error(self):
        """returns the exception varlink error name"""
        return self.args[0].get("error")

    def parameters(self, namespaced=False):
        """returns the exception varlink error parameters"""
        if namespaced:
            return json.loads(
                json.dumps(self.args[0]["parameters"]),
                object_hook=lambda d: SimpleNamespace(**d),
            )
        else:
            return self.args[0].get("parameters")

    def as_dict(self):
        return self.args[0]


class InterfaceNotFound(VarlinkError):
    """The standardized varlink InterfaceNotFound error as a python exception"""

    @classmethod
    def new(cls, message, namespaced=False):
        return cls(
            namespaced and message["parameters"].interface or message["parameters"].get("interface", None)
        )

    def __init__(self, interface):
        VarlinkError.__init__(
            self,
            {
                "error": "org.varlink.service.InterfaceNotFound",
                "parameters": {"interface": interface},
            },
        )


class MethodNotFound(VarlinkError):
    """The standardized varlink MethodNotFound error as a python exception"""

    @classmethod
    def new(cls, message, namespaced=False):
        return cls(namespaced and message["parameters"].method or message["parameters"].get("method", None))

    def __init__(self, method):
        VarlinkError.__init__(
            self,
            {
                "error": "org.varlink.service.MethodNotFound",
                "parameters": {"method": method},
            },
        )


class MethodNotImplemented(VarlinkError):
    """The standardized varlink MethodNotImplemented error as a python exception"""

    @classmethod
    def new(cls, message, namespaced=False):
        return cls(namespaced and message["parameters"].method or message["parameters"].get("method", None))

    def __init__(self, method):
        VarlinkError.__init__(
            self,
            {
                "error": "org.varlink.service.MethodNotImplemented",
                "parameters": {"method": method},
            },
        )


class InvalidParameter(VarlinkError):
    """The standardized varlink InvalidParameter error as a python exception"""

    @classmethod
    def new(cls, message, namespaced=False):
        return cls(
            namespaced and message["parameters"].parameter or message["parameters"].get("parameter", None)
        )

    def __init__(self, name):
        VarlinkError.__init__(
            self,
            {
                "error": "org.varlink.service.InvalidParameter",
                "parameters": {"parameter": name},
            },
        )



"""An implementation of the varlink protocol

See https://www.varlink.org for more information about the varlink protocol and interface definition
files.

For server implementations use the :class:`varlink.Server` class.

For client implementations use the :class:`varlink.Client` class.

For installation and examples, see the GIT repository https://github.com/varlink/python.
or the `source code <_modules/varlink/tests/test_orgexamplemore.html>`_ of
:mod:`varlink.tests.test_orgexamplemore`

"""

import os

if hasattr(os, "fork"):
    __all__ = [
        "Client",
        "ClientInterfaceHandler",
        "SimpleClientInterfaceHandler",
        "Service",
        "RequestHandler",
        "Server",
        "ThreadingServer",
        "ForkingServer",
        "InterfaceNotFound",
        "MethodNotFound",
        "MethodNotImplemented",
        "InvalidParameter",
        "VarlinkEncoder",
        "VarlinkError",
        "Interface",
        "Scanner",
        "get_listen_fd",
    ]
    from .server import ForkingServer
else:
    __all__ = [
        "Client",
        "ClientInterfaceHandler",
        "SimpleClientInterfaceHandler",
        "Service",
        "RequestHandler",
        "Server",
        "ThreadingServer",
        "InterfaceNotFound",
        "MethodNotFound",
        "MethodNotImplemented",
        "InvalidParameter",
        "VarlinkEncoder",
        "VarlinkError",
        "Interface",
        "Scanner",
        "get_listen_fd",
    ]

from .client import Client, ClientInterfaceHandler, SimpleClientInterfaceHandler
from .error import (
    InterfaceNotFound,
    InvalidParameter,
    MethodNotFound,
    MethodNotImplemented,
    VarlinkEncoder,
    VarlinkError,
)
from .scanner import Interface, Scanner
from .server import RequestHandler, Server, Service, ThreadingServer, get_listen_fd


# There are no tests here, so don't try to run anything discovered from
# introspecting the symbols (e.g. FunctionTestCase). Instead, all our
# tests come from within varlink.tests.
def load_tests(loader, tests, pattern):
    import os.path
    import sys
    from fnmatch import fnmatch

    if pattern is None:
        pattern = "test_*.py"

    # top level directory cached on loader instance
    test_dir = os.path.dirname(__file__) + "/tests"
    for fn in os.listdir(test_dir):
        if fnmatch(fn, pattern):
            modname = "varlink.tests." + fn[:-3]
            __import__(modname)
            module = sys.modules[modname]
            tests.addTest(loader.loadTestsFromModule(module))
    return tests



import pytest

import varlink


def test_scanner_1() -> None:
    interface = varlink.Interface("""# Example Varlink service
interface org.example.more

# Enum, returning either start, progress or end
# progress: [0-100]
type State (
  start: ?bool,
  progress: ?int,
  end: ?bool
)

method TestMap(map: [string]string) -> (map: [string](i: int, val: string))

# Returns the same string
method Ping(ping: string) -> (pong: string)

# Dummy progress method
# n: number of progress steps
method TestMore(n: int) -> (state: State)

# Stop serving
method StopServing() -> ()

type ErrorChain (
    description: string,
    caused_by: ?ErrorChain
)

error ActionFailed (reason: ?ErrorChain)
""")
    assert interface.name == "org.example.more"
    assert interface.get_method("Ping") is not None
    assert interface.get_method("TestMore") is not None
    assert interface.get_method("TestMap") is not None
    assert interface.get_method("StopServing") is not None
    assert isinstance(interface.members.get("ActionFailed"), varlink.scanner._Error)
    assert isinstance(interface.members.get("State"), varlink.scanner._Alias)


def test_doubleoption() -> None:
    interface = None
    try:
        interface = varlink.Interface("""
    interface org.example.doubleoption
    method Foo(a: ??string) -> ()
    """)
    except SyntaxError:
        pass

    assert interface is None


def test_complex() -> None:
    interface = varlink.Interface("""
interface org.example.complex

type TypeEnum ( a, b, c )

type TypeFoo (
    bool: bool,
    int: int,
    float: float,
    string: ?string,
    enum: ?[]( foo, bar, baz ),
    type: ?TypeEnum,
    anon: ( foo: bool, bar: int, baz: [](a: int, b: int) ),
    object: object
)

method Foo(a: (b: bool, c: int), foo: TypeFoo) -> (a: [](b: bool, c: int), foo: TypeFoo)

error ErrorFoo (a: (b: bool, c: int), foo: TypeFoo)
""")

    assert interface.name == "org.example.complex"
    assert interface.get_method("Foo") is not None
    assert isinstance(interface.members.get("ErrorFoo"), varlink.scanner._Error)
    assert isinstance(interface.members.get("TypeEnum"), varlink.scanner._Alias)


invalid_interfacenames = [
    "interface .a.b.c\nmethod F()->()",
    "interface com.-example.leadinghyphen\nmethod F()->()",
    "interface com.example-.danglinghyphen-\nmethod F()->()",
    "interface co9.example.number-toplevel\nmethod F()->()",
    "interface 1om.example.number-toplevel\nmethod F()->()",
    "interface ab\nmethod F()->()",
    "interface .a.b.c\nmethod F()->()",
    "interface a.b.c.\nmethod F()->()",
    "interface a..b.c\nmethod F()->()",
    "interface 1.b.c\nmethod F()->()",
    "interface 8a.0.0\nmethod F()->()",
    "interface -a.b.c\nmethod F()->()",
    "interface a.b.c-\nmethod F()->()",
    "interface a.b-.c-\nmethod F()->()",
    "interface a.-b.c-\nmethod F()->()",
    "interface a.-.c\nmethod F()->()",
    "interface a.*.c\nmethod F()->()",
    "interface a.?\nmethod F()->()",
]


@pytest.mark.parametrize("description", invalid_interfacenames)
def test_interfacename_invalid(description) -> None:
    with pytest.raises(SyntaxError):
        varlink.Interface(description)


valid_interfacenames = [
    "interface a.b\nmethod F()->()",
    "interface a.b.c\nmethod F()->()",
    "interface a.1\nmethod F()->()",
    "interface a.0.0\nmethod F()->()",
    "interface org.varlink.service\nmethod F()->()",
    "interface com.example.0example\nmethod F()->()",
    "interface com.example.example-dash\nmethod F()->()",
    "interface xn--lgbbat1ad8j.example.algeria\nmethod F()->()",
    "interface xn--c1yn36f.xn--c1yn36f.xn--c1yn36f\nmethod F()->()",
]


@pytest.mark.parametrize("description", valid_interfacenames)
def test_interfacename_valid(description) -> None:
    assert varlink.Interface(description).name is not None


def test_bad_types() -> None:
    interface = varlink.Interface("""
    interface org.example.testerrors
    type TypeEnum ( a, b, c )
    type TypeDict (dict: [string]string)

    method Foo(param: TypeEnum) -> ()
    method Bar(param: TypeDict) -> ()
    """)
    foo = interface.get_method("Foo")
    with pytest.raises(varlink.InvalidParameter):
        interface.filter_params("test.call", foo.in_type, False, (), {"param": "d"})

    bar = interface.get_method("Bar")
    with pytest.raises(varlink.InvalidParameter):
        interface.filter_params("test.call", bar.in_type, False, (), {"param": {"dict": [1, 2, 3]}})


def test_method_not_found() -> None:
    interface = varlink.Interface("""
    interface org.example.testerrors
    """)

    with pytest.raises(varlink.MethodNotFound):
        interface.get_method("Bar")


def test_struct_errors() -> None:
    missing_colon_after_name = """
    interface org.example.teststruct

    type TypeStruct ( struct ?[](first: int, second: string) )
    """

    with pytest.raises(SyntaxError, match="after 'struct'"):
        varlink.Interface(missing_colon_after_name)

    missing_colon_in_struct = """
    interface org.example.teststruct

    type MyType (
        nullable_array_struct: ?[](first: int, second string)
    )
    method Start() -> (client_id: string)
    """
    with pytest.raises(SyntaxError, match="after 'second'"):
        varlink.Interface(missing_colon_in_struct)


def test_unexpected() -> None:
    invalid = """
    interface org.example.unexpected

    invalid
    """

    with pytest.raises(SyntaxError, match="expected type, method, or error"):
        varlink.Interface(invalid)



# /usr/bin/env python
"""varlink cli tool

Call with:
$ python -m varlink.cli --help

"""

import argparse
import json
import shlex
import socket
import sys

import varlink


def varlink_call(args):
    deli = args.METHOD.rfind(".")
    if deli == -1:
        print("No method found", file=sys.stderr)
        sys.exit(1)

    method = args.METHOD[deli + 1 :]
    interface = args.METHOD[:deli]

    deli = interface.rfind("/")
    if deli != -1:
        address = interface[:deli]
        interface = interface[deli + 1 :]
    else:
        address = None

    def new_client(address):
        if address:
            client = varlink.Client.new_with_address(address)
        else:
            if args.activate:
                client = varlink.Client.new_with_activate(shlex.split(args.activate))
            elif args.bridge:
                client = varlink.Client.new_with_bridge(shlex.split(args.bridge))
            else:
                client = varlink.Client.new_with_resolved_interface(interface, args.resolver)
        return client

    with new_client(address) as client:
        got = False
        try:
            with client.open(interface) as con:
                out = {
                    "method": interface + "." + method,
                    "more": args.more,
                    "parameters": json.loads(args.ARGUMENTS or "{}"),
                }
                con._send_message(json.dumps(out, cls=varlink.VarlinkEncoder).encode("utf-8"))
                more = True
                while more:
                    (message, more) = con._next_varlink_message()
                    if message:
                        print(json.dumps(message, cls=varlink.VarlinkEncoder, indent=2, sort_keys=True))
                        got = True
        except varlink.VarlinkError as e:
            print(e, file=sys.stderr)
        except BrokenPipeError:
            if not got or args.more:
                print("Connection closed")
                sys.exit(1)


def varlink_bridge(args):
    message = b""
    last_interface = None
    con = None
    client = None

    if args.connect:
        client = varlink.Client.new_with_address(args.connect)
        con = client.open_connection()
    elif args.bridge:
        client = varlink.Client.new_with_bridge(shlex.split(args.bridge))
        con = client.open_connection()
    elif args.activate:
        client = varlink.Client.new_with_activate(shlex.split(args.activate))
        con = client.open_connection()
    else:
        print("No --connect or --bridge or --activate")

    if hasattr(sys.stdout, "buffer"):
        stdout = sys.stdout.buffer
    else:
        stdout = sys.stdout

    if hasattr(sys.stdin, "buffer"):
        stdin = sys.stdin.buffer
    else:
        stdin = sys.stdin

    while not sys.stdin.closed:
        c = stdin.read(1)

        if c == b"":
            break

        if c != b"\0":
            message += c
            continue

        req = json.loads(message.decode("utf-8"))

        if not args.connect and not args.activate and not args.bridge:
            if req["method"] == "org.varlink.service.GetInfo":
                req["method"] = "org.varlink.resolver.GetInfo"

            interface_name, _, method_name = req.get("method", "").rpartition(".")

            if req["method"] == "org.varlink.service.GetInterfaceDescription":
                resolving_interface = req["parameters"]["interface"]
            else:
                resolving_interface = interface_name

            if not interface_name or not method_name:
                stdout.write(
                    json.dumps(varlink.InterfaceNotFound(interface_name), cls=varlink.VarlinkEncoder).encode(
                        "utf-8"
                    )
                    + b"\0"
                )
                sys.stdout.flush()
                continue

            if last_interface != resolving_interface:
                if con:
                    if hasattr(con, "shutdown"):
                        con.shutdown(socket.SHUT_RDWR)
                    else:
                        con.close()

                if client:
                    client.cleanup()

                try:
                    client = varlink.Client.new_with_resolved_interface(
                        resolving_interface, resolver_address=args.resolver
                    )
                except varlink.VarlinkError as e:
                    stdout.write(json.dumps(e, cls=varlink.VarlinkEncoder).encode("utf-8") + b"\0")
                    sys.stdout.flush()
                    continue

                con = client.open_connection()
                last_interface = resolving_interface

        if hasattr(con, "send"):
            con.send(message + b"\0")
        else:
            con.write(message + b"\0")

        if req.get("oneway", False):
            continue

        message = b""

        ret_message = b""
        while True:
            c = con.recv(1)

            if c == b"":
                break

            if c != b"\0":
                ret_message += c
                continue

            stdout.write(ret_message + b"\0")

            sys.stdout.flush()

            ret = json.loads(ret_message.decode("utf-8"))
            ret_message = b""

            if not ret.get("continues", False):
                break

            if ret.get("upgraded", False):
                raise NotImplementedError("Bridging upgraded connection not yet supported")


def varlink_help(args):
    deli = args.INTERFACE.rfind("/")
    if deli != -1:
        address = args.INTERFACE[:deli]
        interface_name = args.INTERFACE[deli + 1 :]
        client = varlink.Client.new_with_address(address)
    else:
        interface_name = args.INTERFACE
        if args.activate:
            client = varlink.Client.new_with_activate(shlex.split(args.activate))
        elif args.bridge:
            client = varlink.Client.new_with_bridge(shlex.split(args.bridge))
        else:
            client = varlink.Client.new_with_resolved_interface(interface_name, args.resolver)

    interface = client.get_interface(interface_name)
    del client
    print(interface.description)


def varlink_info(args):
    if args.ADDRESS:
        client = varlink.Client.new_with_address(args.ADDRESS)
    elif args.bridge:
        client = varlink.Client.new_with_bridge(shlex.split(args.bridge))
    elif args.activate:
        client = varlink.Client.new_with_activate(shlex.split(args.activate))
    else:
        print("No ADDRESS or --bridge or --activate")

    client.get_interfaces()
    info = client.info
    del client
    print("Vendor:", info["vendor"])
    print("Product:", info["product"])
    print("Version:", info["version"])
    print("URL:", info["url"])
    print("Interfaces:")
    for i in info["interfaces"]:
        print("  ", i)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(title="commands")
    parser.add_argument("-r", "--resolver", default=None, help="address of the resolver")
    parser.add_argument(
        "-A",
        "--activate",
        default=None,
        help="Service to socket-activate and connect to. "
        "The temporary UNIX socket address is exported as $VARLINK_ADDRESS.",
    )
    parser.add_argument("-b", "--bridge", default=None, help="Command to execute and connect to")

    parser_info = subparsers.add_parser("info", help="Print information about a service")
    parser_info.add_argument("ADDRESS", nargs="?")
    parser_info.set_defaults(func=varlink_info)

    parser_help = subparsers.add_parser("help", help="Print interface description or service information")
    parser_help.add_argument("INTERFACE")
    parser_help.set_defaults(func=varlink_help)

    parser_bridge = subparsers.add_parser(
        "bridge", help="Bridge varlink messages to services on this machine"
    )
    parser_bridge.add_argument(
        "-c",
        "--connect",
        default=None,
        help="Optional varlink address to connect to, without using the resolver.",
    )
    parser_bridge.set_defaults(func=varlink_bridge)

    parser_call = subparsers.add_parser("call", help="Call a method")
    parser_call.add_argument(
        "-m", "--more", action="store_true", help="wait for multiple method returns if supported"
    )
    parser_call.add_argument("METHOD")
    parser_call.add_argument("ARGUMENTS", nargs="?", default="")
    parser_call.set_defaults(func=varlink_call)

    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.print_usage(sys.stderr)
        sys.exit(1)

    args.func(args)



import json
import unittest

import varlink


class OneMessageClientHandler(varlink.ClientInterfaceHandler):
    def __init__(self, interface, namespaced, next_message):
        # No interface but we do not use them
        super().__init__(interface, namespaced)
        self.next_message = next_message

    def _next_message(self):
        yield self.next_message


class TestError(unittest.TestCase):
    def test_pack_unpack(self):
        dummy_if = varlink.Interface("interface org.example.dummy")
        for error in [
            varlink.InterfaceNotFound("org.varlink.notfound"),
            varlink.MethodNotFound("Method"),
            varlink.MethodNotImplemented("Abstract"),
            varlink.InvalidParameter("Struct.param"),
        ]:
            for namespaced in (True, False):
                with self.subTest(error=error, namespaced=namespaced):
                    # encode error
                    encoded = json.dumps(error, cls=varlink.VarlinkEncoder)

                    # Emulates the client receiving an error
                    handler = OneMessageClientHandler(dummy_if, namespaced, encoded)
                    with self.assertRaises(error.__class__):
                        handler._next_varlink_message()



import unittest

import varlink
from varlink import mock

types = """
type MyPersonalType (
    foo: string,
    bar: string
)
"""


class Service:
    # TODO: these type annotations are wrong as Python type hints, but are used
    # by varlink.mock.generate_callable_interface
    def Test1(self, param1: int) -> str:  # type: ignore
        """return test: MyPersonalType"""
        return {
            "test": {
                "foo": "bim",
                "bar": "boom",
            },
        }  # type: ignore

    def Test2(self, param1: str = "test") -> None:
        pass

    def Test3(self, param1: str) -> str:  # type: ignore
        """return test"""
        return {"test": param1}  # type: ignore


class TestMockMechanisms(unittest.TestCase):
    @mock.mockedservice(fake_service=Service, fake_types=types, name="org.service.com", address="unix:@foo")
    def test_init(self):
        with varlink.Client("unix:@foo") as client:
            connection = client.open("org.service.com")
            self.assertEqual(connection.Test1(param1=1)["test"]["bar"], "boom")
            self.assertEqual(connection.Test3(param1="foo")["test"], "foo")


class TestMockUtilities(unittest.TestCase):
    def test_cast_type(self):
        self.assertEqual(mock.cast_type("<class 'int'>"), "int")
        self.assertEqual(mock.cast_type("<class 'str'>"), "string")

    def test_get_ignored(self):
        expected_ignore = dir(mock.MockedService)
        ignored = mock.get_ignored()
        self.assertEqual(expected_ignore, ignored)

    def test_get_attributs(self):
        service = Service()
        attributs = mock.get_interface_attributs(service, mock.get_ignored())
        expected_result = {"callables": ["Test1", "Test2", "Test3"], "others": []}
        self.assertEqual(attributs, expected_result)

    def test_generate_callable_interface(self):
        service = Service()
        generated_itf = mock.generate_callable_interface(service, "Test1")
        self.assertEqual(generated_itf, "method Test1(param1: int) -> (test: MyPersonalType)")
        generated_itf = mock.generate_callable_interface(service, "Test3")
        self.assertEqual(generated_itf, "method Test3(param1: string) -> (test: string)")



#!/usr/bin/env python

"""Test custom JSON encoding by subclassing VarlinkEncoder and
passing it to a server Service.
"""

import dataclasses
import os
import threading
import time
import typing
import unittest

import varlink
import varlink.error


@dataclasses.dataclass
class Shipment:
    name: str
    description: str
    size: int
    weight: typing.Optional[int] = None


@dataclasses.dataclass
class Order:
    shipments: list[Shipment]
    order_num: int
    customer: str


@dataclasses.dataclass
class GetOrderResult:
    order: Order


class VDCEncoder(varlink.error.VarlinkEncoder):
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        return super().default(obj)


service = varlink.Service(
    vendor="Varlink",
    product="Varlink Encoding Example",
    version="1",
    url="http://varlink.org",
    interface_dir=os.path.dirname(__file__),
    json_encoder_cls=VDCEncoder,
)


class ServiceRequestHandler(varlink.RequestHandler):
    service = service


@service.interface("org.example.encoding")
class EncodingExample:
    sleep_duration = 1

    def Ping(self, ping):
        return {"pong": ping}

    def GetOrder(self, num):
        order = Order(
            shipments=[
                Shipment(
                    name="Furniture",
                    description="Furniture by Ferb",
                    size=1000,
                    weight=400,
                ),
                Shipment(
                    name="Electronics",
                    description="Electronics by Earl",
                    size=588,
                ),
            ],
            order_num=num,
            customer="Joe's Discount Store",
        )
        return GetOrderResult(order=order)


class TestService(unittest.TestCase):
    address = "tcp:127.0.0.1:23451"

    @classmethod
    def setUpClass(cls):
        cls._server = varlink.ThreadingServer(cls.address, ServiceRequestHandler)
        cls._server_thread = threading.Thread(target=cls._server.serve_forever)
        cls._server_thread.start()
        time.sleep(0.1)

    @classmethod
    def tearDownClass(cls):
        cls._server.shutdown()
        cls._server.server_close()
        cls._server_thread.join()

    def test_ping(self):
        client = varlink.Client.new_with_address(self.address)
        with client.open("org.example.encoding") as conn:
            response = conn.Ping("Foo")
        self.assertEqual(response["pong"], "Foo")

    def test_get_order(self):
        client = varlink.Client.new_with_address(self.address)
        with client.open("org.example.encoding") as conn:
            response = conn.GetOrder(4638547)
        # response will be a dict represenation of GetOrderResult
        self.assertIn("order", response)
        order = response["order"]
        self.assertEqual(order.get("order_num"), 4638547)
        self.assertEqual(order.get("customer"), "Joe's Discount Store")
        self.assertEqual(len(order.get("shipments", [])), 2)
        shipment1 = order["shipments"][0]
        self.assertEqual(shipment1.get("name"), "Furniture")
        self.assertIsNotNone(shipment1.get("weight"))
        shipment2 = order["shipments"][1]
        self.assertEqual(shipment2.get("name"), "Electronics")
        self.assertIsNone(shipment2.get("weight"))



import re
from collections import OrderedDict
from collections.abc import Mapping, Set
from types import SimpleNamespace

from .error import InvalidParameter, MethodNotFound


class Scanner:
    """Class for scanning a varlink interface definition."""

    def __init__(self, string):
        # whitespace or comment
        wsc = r"([ \t\n]|#.*$)"
        self.whitespace = re.compile(rf"{wsc}+", re.ASCII | re.MULTILINE)
        self.docstring = re.compile(r"(?:.?)+#(.*)(?:\n|\r\n)")
        # FIXME: nested ()
        self.method_signature = re.compile(
            rf"{wsc}*(\(.*?\)){wsc}*->{wsc}*(\(.*?\))",
            re.ASCII | re.MULTILINE,
        )

        self.keyword_pattern = re.compile(
            r"\b[a-z]+\b|[:,(){}]|->|\[\]|\?|\[string\]\(\)|\[string\]", re.ASCII
        )
        self.patterns = {
            "interface-name": re.compile(
                r"[A-Za-z]([A-Za-z])*([.][A-Za-z0-9]([-]*[A-Za-z0-9])*)+|xn--([0-9a-z])*([.][A-Za-z0-9]([-]*[A-Za-z0-9])*)+"
            ),
            "member-name": re.compile(r"\b[A-Z][A-Za-z0-9]*\b", re.ASCII),
            "identifier": re.compile(r"\b[A-Za-z]([_]?[A-Za-z0-9])*\b", re.ASCII),
        }

        self.string = string
        self.pos = 0
        self.current_doc = ""

    def get(self, expected):
        if m := self.whitespace.match(self.string, self.pos):
            doc = self.docstring.findall(self.string[m.start() : m.end()])
            if len(doc):
                try:
                    self.current_doc += "\n".join(doc)
                except UnicodeError:
                    self.current_doc += "\n".join([el.decode("utf-8") for el in doc])
            self.pos = m.end()

        if pattern := self.patterns.get(expected):
            if m := pattern.match(self.string, self.pos):
                self.pos = m.end()
                return m.group(0)
        else:
            m = self.keyword_pattern.match(self.string, self.pos)
            if m and m.group(0) == expected:
                self.pos = m.end()
                return True

    def expect(self, expected):
        if value := self.get(expected):
            return value
        raise SyntaxError(f"expected '{expected}'")

    def end(self):
        if m := self.whitespace.match(self.string, self.pos):
            doc = self.docstring.findall(self.string[m.start() : m.end()])
            if len(doc):
                try:
                    self.current_doc += "\n".join(doc)
                except UnicodeError:
                    self.current_doc += "\n".join([el.decode("utf-8") for el in doc])
            self.pos = m.end()

        return self.pos >= len(self.string)

    def read_type(self, lastmaybe=False):
        if self.get("?"):
            if lastmaybe:
                raise SyntaxError("double '??'")
            return _Maybe(self.read_type(lastmaybe=True))

        if self.get("[string]()"):
            return set()

        if self.get("[string]"):
            return _Dict(self.read_type())

        if self.get("[]"):
            return _Array(self.read_type())

        if self.get("object"):
            return _Object()

        if self.get("bool"):
            t = False
        elif self.get("int"):
            t = 0
        elif self.get("float"):
            t = 0.0
        elif self.get("string"):
            t = ""
        else:
            name = self.get("member-name")
            if name:
                t = _CustomType(name)
            else:
                t = self.read_struct()

        return t

    def read_struct(self):
        _isenum = None
        self.expect("(")
        fields = OrderedDict()
        if not self.get(")"):
            while True:
                name = self.expect("identifier")
                if _isenum is None:
                    if self.get(":"):
                        _isenum = False
                        fields[name] = self.read_type()
                        if not self.get(","):
                            break
                        continue
                    elif self.get(","):
                        _isenum = True
                        fields[name] = True
                        continue
                    else:
                        raise SyntaxError(f"after '{name}'")
                elif not _isenum:
                    try:
                        self.expect(":")
                        fields[name] = self.read_type()
                    except SyntaxError as e:
                        raise SyntaxError(f"after '{name}': {e}")
                else:
                    fields[name] = True

                if not self.get(","):
                    break
            self.expect(")")
        if _isenum:
            return _Enum(fields.keys())
        else:
            return _Struct(fields)

    def read_member(self):
        if self.get("type"):
            try:
                _name = self.expect("member-name")
            except SyntaxError:
                m = self.whitespace.match(self.string, self.pos)
                if m:
                    start = m.end()
                else:
                    start = self.pos
                m = self.whitespace.search(self.string, start)
                if m:
                    stop = m.start()
                else:
                    stop = start

                raise SyntaxError(f"'{self.string[start:stop]}' not a valid type name.")
            try:
                _type = self.read_type()
            except SyntaxError as e:
                raise SyntaxError(f"in '{_name}': {e}")
            doc = self.current_doc
            self.current_doc = ""
            return _Alias(_name, _type, doc)
        elif self.get("method"):
            name = self.expect("member-name")
            # FIXME
            sig = self.method_signature.match(self.string, self.pos)
            if sig:
                sig = name + sig.group(0)
            in_type = self.read_struct()
            self.expect("->")
            out_type = self.read_struct()
            doc = self.current_doc
            self.current_doc = ""
            return _Method(name, in_type, out_type, sig, doc)
        elif self.get("error"):
            doc = self.current_doc
            self.current_doc = ""
            return _Error(self.expect("member-name"), self.read_type(), doc)
        else:
            raise SyntaxError("expected type, method, or error")


class _Object:
    pass


class _Struct:
    def __init__(self, fields):
        self.fields = OrderedDict(fields)


class _Enum:
    def __init__(self, fields):
        self.fields = fields


class _Array:
    def __init__(self, element_type):
        self.element_type = element_type


class _Maybe:
    def __init__(self, element_type):
        self.element_type = element_type


class _Dict:
    def __init__(self, element_type):
        self.element_type = element_type


class _CustomType:
    def __init__(self, name):
        self.name = name


class _Alias:
    def __init__(self, name, varlink_type, doc=None):
        self.name = name
        self.type = varlink_type
        self.doc = doc


class _Method:
    def __init__(self, name, in_type, out_type, _signature, doc=None):
        self.name = name
        self.in_type = in_type
        self.out_type = out_type
        self.signature = _signature
        self.doc = doc


class _Error:
    def __init__(self, name, varlink_type, doc=None):
        self.name = name
        self.type = varlink_type
        self.doc = doc


class Interface:
    """Class for a parsed varlink interface definition."""

    def __init__(self, description):
        """description -- description string in varlink interface definition language"""
        self.description = description

        scanner = Scanner(description)
        scanner.expect("interface")
        self.name = scanner.expect("interface-name")
        self.doc = scanner.current_doc
        scanner.current_doc = ""
        self.members = OrderedDict()
        while not scanner.end():
            member = scanner.read_member()
            self.members[member.name] = member

    def get_description(self):
        """return the description string in varlink interface definition language"""
        return self.description

    def get_method(self, name):
        method = self.members.get(name)
        if method and isinstance(method, _Method):
            return method
        raise MethodNotFound(name)

    def filter_params(self, parent_name, varlink_type, _namespaced, args, kwargs):
        # print("filter_params", type(varlink_type), repr(varlink_type), args, kwargs, type(args))

        if isinstance(varlink_type, _Maybe):
            if args is None:
                return None
            return self.filter_params(parent_name, varlink_type.element_type, _namespaced, args, kwargs)

        if isinstance(varlink_type, _Dict):
            if args is None:
                return {}

            if isinstance(args, Mapping):
                for k, v in args.items():
                    args[k] = self.filter_params(
                        parent_name + "[" + k + "]", varlink_type.element_type, _namespaced, v, None
                    )
                return args
            else:
                raise InvalidParameter(parent_name)

        if isinstance(varlink_type, _CustomType):
            # print("CustomType", varlink_type.name)
            return self.filter_params(
                parent_name, self.members.get(varlink_type.name), _namespaced, args, kwargs
            )

        if isinstance(varlink_type, _Alias):
            # print("Alias", varlink_type.name)
            return self.filter_params(parent_name, varlink_type.type, _namespaced, args, kwargs)

        if isinstance(varlink_type, _Object):
            return args

        if isinstance(varlink_type, _Enum) and isinstance(args, str):
            if args not in varlink_type.fields:
                raise InvalidParameter(parent_name)
            return args

        if isinstance(varlink_type, _Array):
            if args is None:
                return []

            return [
                self.filter_params(parent_name + "[]", varlink_type.element_type, _namespaced, x, None)
                for x in args
            ]

        if isinstance(varlink_type, Set):
            # print("Returned set:", set(args))
            return set(args)

        if isinstance(varlink_type, str) and isinstance(args, str):
            return args

        if isinstance(varlink_type, float) and (isinstance(args, float) or isinstance(args, int)):
            # print("Returned float:", args)
            return float(args)

        if isinstance(varlink_type, bool) and isinstance(args, bool):
            # print("Returned bool:", args)
            return args

        if isinstance(varlink_type, int) and (isinstance(args, float) or isinstance(args, int)):
            # print("Returned int:", args)
            if isinstance(args, float):
                return int(args + 0.5)
            return int(args)

        if not isinstance(varlink_type, _Struct):
            raise InvalidParameter(parent_name)
            # SyntaxError(f"Expected type {type(varlink_type)}, got {type(args)} with value '{args}'"

        if _namespaced:
            out = SimpleNamespace()
        else:
            out = {}

        varlink_struct = None
        if not isinstance(args, tuple):
            varlink_struct = args
            args = None

        for name in varlink_type.fields:
            if isinstance(args, tuple):
                if args:
                    val = args[0]
                    if len(args) > 1:
                        args = args[1:]
                    else:
                        args = None
                    ret = self.filter_params(
                        parent_name + "." + name, varlink_type.fields[name], _namespaced, val, None
                    )
                    if ret is not None:
                        # print("SetOUT:", name)
                        if _namespaced:
                            setattr(out, name, ret)
                        else:
                            out[name] = ret
                    continue
                else:
                    if name in kwargs:
                        ret = self.filter_params(
                            parent_name + "." + name,
                            varlink_type.fields[name],
                            _namespaced,
                            kwargs[name],
                            None,
                        )
                        if ret is not None:
                            # print("SetOUT:", name)
                            if _namespaced:
                                setattr(out, name, ret)
                            else:
                                out[name] = ret
                        continue

            if varlink_struct:
                if isinstance(varlink_struct, Mapping):
                    if name not in varlink_struct:
                        continue

                    val = varlink_struct[name]
                    ret = self.filter_params(
                        parent_name + "." + name, varlink_type.fields[name], _namespaced, val, None
                    )
                    if ret is not None:
                        # print("SetOUT:", name)
                        if _namespaced:
                            setattr(out, name, ret)
                        else:
                            out[name] = ret
                elif hasattr(varlink_struct, name):
                    val = getattr(varlink_struct, name)
                    ret = self.filter_params(
                        parent_name + "." + name, varlink_type.fields[name], _namespaced, val, None
                    )
                    if ret is not None:
                        # print("SetOUT:", name)
                        if _namespaced:
                            setattr(out, name, ret)
                        else:
                            out[name] = ret
                else:
                    continue

        return out
